#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 14:05:10 2018

@author: yoelr
"""
from . import _Q
import numpy as np
from ._utils import property_array, PropertyFactory, DisplayUnits, \
                    tuple_array, fraction, Sink, Source, MissingStream
from ._flowsheet import find
from ._species import Species, WorkingSpecies
from ._exceptions import SolverError, EquilibriumError, DimensionError
from ._equilibrium import Dortmund, VLE, BubblePoint, DewPoint


__all__ = ('Stream',)

# %% TODOs

# TODO: add material property interphase when using Cape Open package

# %% Functions

def nonzero_species(species, flow):
    index_ = []
    IDs_ = []
    IDs = species._IDs
    for i in species._index:
        if flow[i] != 0:
            index_.append(i)
            IDs_.append(IDs[i])
    return index_, IDs_

def _print_helpdata(helpdata):
    """Print help data."""
    # Only one helplist, print a nice string
    if isinstance(helpdata[0], str):
        propID, description, dependency, units, datatype = helpdata
        if dependency == 'TP':
            dependency = 'as a function of T and P '
        elif dependency == 'T':
            dependency = 'as a function of T '
        elif dependency == 'P':
            dependency = 'as a function of P '
        print(f"{propID}: [{datatype}] {description.capitalize()} "
              "{dependency}({units}).")

    # Many helpdata, print all the nice strings
    else:
        for i in helpdata:
            _print_helpdata(i)


# %% Units of measure

# Biosteam units of measure
units_of_measure = dict(cost='USD/hr',
                        MW='g/mol',
                        mass='kg/hr',
                        mol='kmol/hr',
                        vol='m^3/hr',
                        massnet='kg/hr',
                        molnet='kmol/hr',
                        volnet='m^3/hr',
                        massfrac='kg/kg',
                        molfrac='kmol/kmol',
                        volfrac='m^3/m^3',
                        T='K',
                        P='Pa',
                        H='kJ/hr',
                        S='kJ/hr',
                        G='kJ/hr',
                        U='kJ/hr',
                        A='kJ/hr',
                        Hf='kJ/hr',
                        C='kJ/K/hr',
                        Vm='m^3/mol',
                        Cpm='J/mol/K',
                        Cp='J/g/K',
                        rho='kg/m^3',
                        rhom='mol/m^3',
                        nu='m^2/s',
                        mu='Pa*s',
                        sigma='N/m',
                        k='W/m/K',
                        alpha='m^2/s')

mol_flow_dim = _Q(0, units_of_measure['mol']).dimensionality
mass_flow_dim = _Q(0, units_of_measure['mass']).dimensionality
vol_flow_dim = _Q(0, units_of_measure['vol']).dimensionality

# %% Flow properties

@PropertyFactory
def MassFlow(self):
    """Mass flow (kg/hr)."""
    return self.data[0][0] * self.data[1] # mol[0] * MW

@MassFlow.setter
def MassFlow(self, value):
    self.data[0][0] = value/self.data[1] # mol[0] = value/MW

@PropertyFactory    
def VolumetricFlow(self):
    """Volumetric flow (m^3/hr)."""
    stream, mol = self.data
    m = mol[0]
    if m:
        c = self.name # c = compound
        c.T = stream.T
        c.P = stream.P
        c.phase = stream._phase
        return c.Vm * m * 1000
    else:
        return 0.

@VolumetricFlow.setter
def VolumetricFlow(self, value):
    stream, mol = self.data
    if value:
        c = self.name # c = compound
        c.T = stream.T
        c.P = stream.P
        c.phase = stream._phase
        mol[0] = value/(c.Vm * 1000)
    else:
        mol[0] = 0.

phases = ('s', 'l', 'L', 'g')
phase_index = dict(zip(phases, (0, 1, 2, 3)))

def flow(fget):
    def fset(self, value):
        if fget(self) is not value: raise AttributeError(f"can't set attribute")
    return property(fget, fset)


# %% Stream classes

class metaStream(type):
    """Metaclass for Stream."""
    @property
    def species(cls):
        """[Species] Contains pure component thermodynamic properties for computing overall properties of Stream instances."""
        return cls._cls_species
    @species.setter
    def species(cls, species):
        # Set Species object and related parameters
        if isinstance(species, Species):
            Stream._cls_species = WorkingSpecies(species)
        elif isinstance(species, WorkingSpecies):
            Stream._cls_species = species
        else: raise ValueError('must pass a Species object')
    _species = species
    @property
    def indices(self):
        return self._cls_species.indices
    @property
    def index(self):
        return self._cls_species.index
    @property
    def MW(cls):
        return cls._MW

class Stream(metaclass=metaStream):
    """Create a Stream object that defines material flow rates along its thermodynamic state. Thermodynamic and transport properties of a stream are readily available. Ideal mixture is assumed for stream properties and excess thermodynamic energies are neglected as a simplifying assumption for low pressure processes.

**Parameters**

    **ID:** [str] A unique identification. If ID is an empty string (i.e. '' ), a default ID will be chosen. If ID is None, stream will not be registered in flowsheet.

    **flow:** [tuple] All flow rates corresponding to `species`.

    **species:** tuple[str] or [Species] Species corresponding to `flow`. If empty, assume same species as class.

    **units:** [str] The units of the flow rates (only mass, molar, and volumetric flow rates are valid)

    **phase:** [str] 'l' for liquid, 'g' for gas, or 's' for solid.

    **T:** [float] Temperature (K).

    **P:** [float] Pressure (Pa).

    **price:** [float] (USD/kg).
    
    ****flow_pairs:** Compound-flow pairs

**Class Properties**

    **species** = Species(): [Species] Contains pure component thermodynamic properties for computing overall properties of Stream instances.

**Examples**

    Before making a stream, set the species using a Species object:

    .. code-block:: python

       >>> # Set Species object
       >>> Stream.species = Species('Ethanol', 'Water') 

    Stream objects may be created a variety of ways:

    .. code-block:: python

       >>> # Create a stream specifying compound and flow rate pairs:
       >>> s1 = Stream(ID='s1', Water=2)
       >>> s1.show()
       Stream: s1
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  2

       >>> # Create a stream assuming same order as given in species:
       >>> s2 = Stream(ID='s2', flow=(1, 2))
       >>> s2.show()
       Stream: s2
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  1
                        Water    2

       >>> # Create a stream passing flow rate units, phase, temperature and pressure:
       >>> s3 = Stream(ID='s3', flow=(0.278, 0.556), units='mol/s', phase='g', T=400, P=101325)
       >>> s3.show()
       Stream: s3
        phase: 'g', T: 400 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  1.
                        Water    2.

       >>> # The working units do not change
       >>> s3.mol 
       array([1., 2.])

    .. Warning:: Stream objects do not automatically calculate thermodynamic equilibrium. They simply assume the given phase, temperature and pressure are correct. To find equilibrium, use the VLE or LLE method.

    Use the `show` method to print all specifications with desired units:

    .. code-block:: python

       >>> # Temperature in degree Celsius
       >>> s2.show(T='degC')
       Stream: s2 
        phase: 'l', T: 25. degC, P: 101325 Pa
        flow (kmol/hr): Ethanol  1
                        Water    2
       
       >>> # Flow in kg/hr
       >>> s2.show(flow='kg/hr')
       Stream: s2 
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kg/hr): Ethanol  46.1
                      Water    36

       >>> # Flow in fractions
       >>> s2.show(fraction=True)
       Stream: s2
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: Ethanol  0.333
              Water    0.667
              net      3 kmol/hr

    Flow rates are stored internally as an array in the ‘mol’ attribute.

    .. code-block:: python

       >>> # Set Water flow rate
       >>> s2.mol[1] = 18
       >>> s2.mol # kmol/hr
       array([1, 18])

    Mass and volumetric flow rates are also available as property_arrays of the molar flow rate. As such, they are always up to date with the molar flow rate and altering them also alters the molar flow rate:

    .. code-block:: python

       >>> # Altering mass or volumetric flows alters the molar flow
       >>> s2.vol # m^3/hr
       property_array([0.059, 0.036])
       >>> s2.vol[:] = [1, 1]
       >>> s2.mol
       array([17.06 , 55.343])
       >>> # Values are always up to date with the molar flow
       >>> s2.mol[:] = [1, 2]
       >>> s2.vol
       property_array([0.059, 0.036])

    .. Note::

       property_array objects are significantly slower than array objects. This is because flow rate data is internally stored as molar flow rates. Also, property_array objects are arrays of python objects, which add overhead over the C implementation of numpy. Whenever possible, use the array to manage flow rates. 

    Some thermodynamic/material properties, including enthalpy and heat capacity, are dependent only on temperature:

    .. code-block:: python
    
       >>> s2.T = 298.15
       >>> s2.Cp # Heat capacity (kJ/(kg-K))
       3.200382054794244
       >>> s2.H  # Enthalpy (kJ/hr)
       0.0
       
       >>> # Change of temperature
       >>> s2.T = 310
       >>> s2.H
       3140.9389548625936
       >>> s2.Cp
       3.2590872180092956
       
    .. Note:: Thermodynamic energies are relative to 25 degC and 1 atm.
    
    Some thermodynamic/material properties, including volume and density, are dependent on both temperature and pressure:

    .. code-block:: python
    
       >>> s1.volnet # Volumetric flow rate (m^3/hr)
       0.036138079740245625
       >>> s1.rho # Density (kg/m^3)
       997.0247522552814
      
       >>> # Change of pressure
       >>> s1.P *= 4
       >>> s1.volnet
       0.036136231155141196
       >>> s1.rho
       997.0757560552587
      
       >>> # Change of temperature
       >>> s1.T += 30
       >>> s1.volnet
       0.03663597700908213
       >>> s1.rho
       983.4747955832584

    A dictionary of available stream properties and respective units of measure is available in `Stream.units`. You may also find it useful to use the `help` method to search for a property:
        
    .. code-block:: python

       >>> Stream.help('conductivity')
       k: [float] Thermal conductivity as a function of T and P (W/m/K).

    """
    
    # [dict] Units of measure for material properties (class attribute). 
    units = units_of_measure

    # Information regarding properties
    _prop_info = (
        # ID         # Description               # Dependency # Units      # Type
        ('T',        'temperature',              '',          'K',         'float'),
        ('H',        'enthalpy',                 'T',         'kJ/hr',     'float'),
        ('S',        'entropy',                  'TP',        'kJ/hr',     'float'),
        ('G',        'Gibbs free energy',        'TP',        'kJ/hr',     'float'),
        ('U',        'interal energy',           'TP',        'kJ/hr',     'float'),
        ('A',        'Helmholtz free energy',    'TP',        'kJ/hr',     'float'),
        ('Hf',       'enthalpy of formation',    '',          'kJ/hr',     'float'),
        ('P',        'pressure',                 '',          'Pa',        'float'),
        ('Cpm',      'molar heat capacity',      'T',         'J/mol/K',   'float'),
        ('Cp',       'specific heat capacity',   'T',         'J/kg/K',    'float'),
        ('Vm',       'molar volume',             'TP',        'm^3/mol',   'float'),
        ('rho',      'density',                  'TP',        'kg/m^3',    'float'),
        ('rhom',     'molar density',            'TP',        'mol/m^3',   'float'),
        ('nu',       'kinematic viscosity',      'TP',        'm^2/s',     'float'),
        ('mu',       'hydraulic viscosity',      'TP',        'Pa*s',      'float'),
        ('sigma',    'surface tension',          'T',         'N/m',       'float'),
        ('k',        'thermal conductivity',     'TP',        'W/m/K',     'float'),
        ('alpha',    'thermal diffusivity',      'TP',        'm^2/s',     'float'),
        ('Pr',       'Prantl number',            'TP',        "''",        'float'),
        ('mass',     'mass flow rates',          '',          'kg/hr',     'ndarray'),
        ('mol',      'molar flow rates',         '',          'kmol/hr',   'ndarray'),
        ('vol',      'volumetric flow rates',    'TP',        'm^3/hr',    'ndarray'),
        ('massnet',  'net mass flow rate',       '',          'kg/hr',     'float'),
        ('molnet',   'net molar flow rate',      '',          'kmol/hr',   'float'),
        ('volnet',   'net volumetric flow rate', 'TP',        'm^3/hr',    'float'),
        ('massfrac', 'mass fractions',           '',          'kg/kg',     'ndarray'),
        ('molfrac',  'molar fractions',          '',          'kmol/kmol', 'ndarray'),
        ('volfrac',  'volumetric fractions',     'TP',        'm^3/m^3',   'ndarray'))

    __slots__ = ('T', 'P', '_mol', '_mass', '_vol', 'price', '_ID', '_link',
                 '_species', '_sink', '_source', '_dew_point',
                 '_bubble_point', '_gamma', '_phase', '_lL_split_cached',
                 '__weakref__', '_source_link', '_VLE')

    line = 'Stream'

    ### Class attributes for working species ###    
    _cls_species = _MW = None
    
    # [list] Default starting letter and current number for ID (class attribute)
    _default_ID = ['d', 1]
    
    #: [bool] If True, approximate energy balance. False otherwise.
    lazy_energy_balance = True

    #: [DisplayUnits] Units of measure for IPython display
    display_units = DisplayUnits(T='K', P='Pa',
                                 flow=('kmol/hr', 'kg/hr', 'm3/hr'),
                                 fraction=False)

    def __init__(self, ID='', flow=(), species=(), units='kmol/hr',
                 phase='l', T=298.15, P=101325, *, price=0, **flow_pairs):
        # Get species and set species information
        if isinstance(species, Species):
            self._species = WorkingSpecies(species)
            species = ()
        elif isinstance(species, WorkingSpecies):
            self._species = species
            species = ()
        else: 
            assert self._cls_species, 'must define Stream.species first'
            self._species = self._cls_species
        self._link = self._ID = self._sink = self._source = None
        self._source_link = self
        self.phase = phase
        self.T = T  #: [float] Temperature (K)
        self.P = P  #: [float] Pressure (Pa)
        self.price = price  #: Price of stream (USD/kg)
        
        # Initialize flows
        self._setflows(flow, species, flow_pairs)
        mol = self._mol
        MW = self._species._MW
        mass = [] # Mass flow rates
        vol = [] # Volumetric flow rates    
        cmps = self._species._compounds
        for i in self._species._index:
            mol_i = mol[i:i+1]
            s = cmps[i]
            mass.append(MassFlow(s.ID, (mol_i, MW[i])))
            vol.append(VolumetricFlow(s, (self, mol_i)))
        self._mass = property_array(mass)
        self._vol = property_array(vol)
        if units == 'kmol/hr': pass
        if units == 'kg/hr': self._mass[:] = mol
        elif units == 'm3/hr': self._vol[:] = mol
        else:
            q = _Q(mol, units)
            dim = q.dimensionality
            if dim == mol_flow_dim:
                self._mol[:] = q.to('kmol/hr').magnitude
            elif dim == mass_flow_dim:
                self._mass[:] = q.to('kg/hr').magnitude
            elif dim == vol_flow_dim:
                self._vol[:] = q.to('m3/hr').magnitude
            else:
                raise DimensionError(f"dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'")
        self.ID = ID
        self._gamma = gamma = Dortmund()
        self._bubble_point = BubblePoint(gamma)
        self._dew_point = DewPoint(gamma)

    def setflow(self, flow=(), species=(), units='kmol/hr', inplace='', **flow_pairs):
        """Set `flow` rates according to the `species` order and `flow_pairs`. `inplace` can be any operation that can be performed in place (e.g. +, -, *, /, |, **, etc.)."""
        species = (*species, *flow_pairs.keys())
        flow = (*flow, *flow_pairs.values())
        index = self.indices(species) if species else ... 
        q = _Q(flow, units)
        dim = q.dimensionality
        if dim == mol_flow_dim:
            exec(f"self._mol[index] {inplace}= q.to('kmol/hr').magnitude", locals())
        elif dim == mass_flow_dim:
            exec(f"self._mass[index] {inplace}= q.to('kg/hr').magnitude", locals())
        elif dim == vol_flow_dim:
            exec(f"self._vol[index] {inplace}= q.to('m3/hr').magnitude", locals())
        else:
            raise DimensionError(f"dimensions for flow units must be in molar, "
                                 f"mass or volumetric flow rates, not '{dim}'")
    
    def getflow(self, *species, units='kmol/hr'):
        """Get flow rates of species in given units."""
        index = self.indices(species) if species else ...
        q = _Q(1, units)
        dim = q.dimensionality
        if dim == mol_flow_dim:
            return self._mol[index]*q.to('kmol/hr').magnitude
        elif dim == mass_flow_dim:
            return self._mass[index]*q.to('kg/hr').magnitude
        elif dim == vol_flow_dim:
            return self._vol[index]*q.to('m3/hr').magnitude
        else:
            raise DimensionError(f"dimensions for flow units must be in molar, "
                                 f"mass or volumetric flow rates, not '{dim}'")
    
    def _setflows(self, flow, species, flow_pairs):
        """Initialize molar flow rates according to the species order, and flow_pairs. Instance species do not change."""
        flowlen = len(flow)
        specieslen = len(species)
        if flowlen:
            if flow_pairs:
                raise ValueError('cannot specify both "flow" and '
                                 '"flow_pairs" when species is passed')
            elif flowlen == specieslen:
                self._mol = self._species.array(species, flow)
            elif (not specieslen) and (flowlen == self._species._N):
                self._mol = np.array(flow, float)
            else:
                raise ValueError('length of flow rates must be equal to '
                                 'length of species')
        elif flow_pairs:
            self._mol = self._species.array(flow_pairs, [*flow_pairs.values()])
        else:
            self._mol = np.zeros(self.species._N, float)
                
    # Forward pipping
    def __sub__(self, index):
        if isinstance(index, int):
            return Sink(self, index)
        elif isinstance(index, Stream):
            raise TypeError("unsupported operand type(s) for -: "
                            f"'{type(self)}' and '{type(index)}'")
        return index.__rsub__(self)
        
    def __rsub__(self, index):
        if isinstance(index, int):
            return Source(self, index)
        elif isinstance(index, Stream):
            raise TypeError("unsupported operand type(s) for -: "
                            "'{type(self)}' and '{type(index)}'")
        return index.__sub__(self)

    # Backward pipping    
    __pow__ = __sub__
    __rpow__ = __rsub__
    
    @property
    def phase(self):
        """[str] 'l' for liquid, 'g' for gas, or 's' for solid."""
        return self._phase
    @phase.setter
    def phase(self, phase):
        if phase not in ('l', 'g', 's'):
            raise ValueError(f"phase must be either 's', 'l', or 'g'")
        self._phase = phase
    
    @staticmethod
    def proxy(ID, link=MissingStream):
        """Create a Stream object that serves as a proxy for its `link` stream."""
        self = object.__new__(Stream)
        self._source = self._sink = self._ID = None
        self._source_link = MissingStream
        self.ID = ID
        self.price = 0
        if link: self.link = link
        else: self._link = MissingStream
        return self
    
    @property
    def link(self):
        """When another Stream object is set as a link, it will share data with that object."""
        return self._link
    
    @link.setter
    def link(self, stream):
        try:
            if self._source_link is stream._source_link:
                self._link = stream
            elif stream is None:
                self.__init__(ID=self._ID, flow=self._mol,
                              species=self._species,
                              T=self.T, P=self.P, phase=self._phase)
            else:
                self._species = stream._species
                self._mass = stream._mass
                self._mol = stream._mol
                self._vol = stream._vol
                self._dew_point = stream._dew_point
                self._bubble_point = stream._bubble_point
                self._gamma = stream._gamma
                self._source_link = stream._source_link
                self._link = stream
                self.P = stream.P
                self.T = stream.T
                self._phase = stream.phase
        except Exception as Error:
            if isinstance(stream, Stream): raise Error
            else: raise TypeError(f"link must be a Stream object, not a "
                                  "'{type(stream).__name__}' object.")
    
    # def __getattr__(self, key):
    #     if self._link is MissingStream:
    #         if self.source and self.source._ins[0] is not MissingStream:
    #             self.source._link_streams()
    #             try: return object.__getattribute__(self, key)
    #             except: raise AttributeError(f'{repr(self.source)}.ins[0] is missing stream')
    #         raise AttributeError(f'{self}.link is missing stream')
    #     elif not self._source_link:
    #         raise AttributeError(f'{repr(self.link.source)} must be simulated first')
    #     else:
    #         raise AttributeError(f"'{type(self).__name__}' has no attribute '{key}'")
    
    @property
    def species(self):
        """[Species] Contains pure component thermodynamic properties for computing overall properties of Stream instances."""
        return self._species
    
    @property
    def ID(self):
        """Unique identification (str). If set as '', it will choose a default ID."""
        return self._ID

    @ID.setter
    def ID(self, ID):
        if ID == '':
            # Select a default ID
            letter, number = self._default_ID
            self._default_ID[1] += 1
            num = str(number)
            ID = letter + num
            self._ID = ID
            find.stream[ID] = self
        elif ID and ID != self._ID:
            ID = ID.replace(' ', '_')
            ID_words = ID.split('_')
            if not all([word.isalnum() for word in ID_words]):
                raise ValueError('ID cannot have any special characters')
            self._ID = ID
            find.stream[ID] = self

    @property
    def cost(self):
        """Cost of stream (USD/hr)"""
        return self.price*self.massnet

    @property
    def MW(self):
        """Molecular weight of all species (array g/mol):     

        from biosteam import *
        >>> Stream.species = Species('Water', 'Ethanol')
        >>> Stream.MW
        tuple_array([18.01528, 46.06844])
        """
        return self._species._MW.view(tuple_array)

    ### Flow properties ###
    
    @property
    def indices(self):
        """Return flow indices of specified species.

        **Parameters**

             **IDs:** [iterable] species IDs or CAS numbers.

        Example

        Indices by ID:
        
        >>> from biosteam import *
        >>> Stream.species = Species(['Ethanol', 'Water'])
        >>> s1 = Stream()
        >>> s1.indices(['Water', 'Ethanol'])
        [1, 0]

        Indices by CAS number:
        
        >>> s1.indices(['7732-18-5', '64-17-5']):
        [1, 0]

        """
        return self._species.indices
    
    @property
    def index(self):
        """Return index of specified compound.

        **Parameters**

             **ID:** [str] compound ID

        Example

        Index by ID:
        
        >>> from biosteam import *
        >>> Stream.species = Species(['Ethanol', 'Water'])
        >>> s1 = Stream()
        >>> s1.index('Water')
        1

        Indices by CAS number:
        
        >>> s1.index('7732-18-5'):
        1

        """
        return self._species.index
    
    # Molar flow
    @flow
    def mol(self):
        """Array of molar flow rates (kmol/hr):

        >>> from biosteam import *
        >>> Stream.species = Species('Water', 'Ethanol')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.mol
        array([0, 2])
        >>> s1.mol[:] = [1, 2]
        >>> s1.mol
        array([1, 2])
        """
        return self._mol

    @property
    def molfrac(self):
        """Array of molar fractions.

        >>> from biosteam import *
        >>> Stream.species = Species('Water', 'Ethanol')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.molfrac
        tuple_array([0., 1.])
        """
        return fraction(self._mol).view(tuple_array)

    @property
    def molnet(self):
        """Net molar flow rate (kmol/hr)

        >>> from biosteam import *
        >>> Stream.species = Species('Water', 'Ethanol')
        >>> s1 = Stream(Ethanol=2, Water=1)
        >>> s1.molnet
        3
        """
        return self._mol.sum()

    # Mass flow
    @flow
    def mass(self):
        """Array of mass flow rates (kg/hr)

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.mass
        property_array([ 0.   , 36.031])
        """
        return self._mass

    @property
    def massfrac(self):
        """Array of mass fractions.

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.massfrac
        tuple_array([0, 1])
        """
        return fraction(self._mol * self._species._MW).view(tuple_array)

    @property
    def massnet(self):
        """Net mass flow rate (kg/hr)

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.mass
        36.031
        """
        return (self._species._MW * self._mol).sum()

    # Volumetric flow
    @flow
    def vol(self):
        """Array of volumetric flow rates as a function of T and P (m^3/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.vol
        property_array([0.   , 0.036])
        
        """
        return self._vol

    @property
    def volfrac(self):
        """Array of volumetric fractions as a function of T and P.

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.volfrac
        tuple_array([0.0, 1.0], dtype=object)
                """
        return fraction(self._vol).view(tuple_array)
        
    @property
    def volnet(self):
        """Net volumetric flow rate as a function of T and P (m^3/hr).

        from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.volnet
        0.036138079740245625
        """
        return self._species._propflow('Vm', self._mol, self.T, self.P, self._phase)*1000

    ### Energy flows ###

    def H_at(self, mol=None, T=None, P=None, phase=None):
        """Return enthalpy flow rate at given arguments, excluding formation energies (kJ/hr). Arguments with None values default to stream specifications."""
        return self._species._propflow('H', mol or self._mol, T or self.T,
                                       P or self.P, phase or self._phase)

    # Enthalpy
    @property
    def H(self):
        """Enthalpy flow rate as a function of T, excluding formation energies (kJ/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.H # The stream is at the reference state
        0.0
        >>> s1.H = 1000
        >>> s1.T
        304.7888167493753
        >>> s1.H
        999.5728716085112

        .. Note:: The solver reaches a level of tolerance and the new temperature is an approximation.
        """
        return self._species._propflow('H', self._mol, self.T, self.P, self._phase)

    @H.setter
    def H(self, H):
        try:
            if self.lazy_energy_balance:
                self.T += (H - self.H)/self.C    
            else:
                # First approximation
                T_old = self.T
                C = self.C
                self.T += (H - self.H)/C
            
                # Solve enthalpy by iteration
                it = 0
                it2 = 0
                while abs(self.T - T_old) > 0.01:
                    T_old = self.T
                    self.T += (H - self.H)/C
                    if it == 5:
                        it = 0
                        it2 += 1
                        C = self.C
                        if it2 > 10:
                            raise SolverError("could not solve temperature "
                                              "given enthalpy")
                    else: it += 1
        except Exception as Error:
            if (self._mol == 0).all():
                raise ValueError(f"cannot set enthalpy to empty stream")
            else:
                raise Error

    @property
    def Hnet(self):
        """Total enthaly flow rate including heats of formation"""
        return self.Hf + self.H

    @property
    def Hf(self):
        """Heat of formation flow rate (kJ/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.Hf
        -483640.0
        """
        return (self.mol * [i.Hf or 0 for i in self._species._compounds]).sum()

    @property
    def Hc(self):
        """Heat of combustion flow rate (kJ/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.Hc
        -2819966.0
        """
        return (self.mol * [i.Hc or 0 for i in self._species._compounds]).sum()
        
    # Entropy
    @property
    def S(self):
        """Entropy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.S # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.S
        16.503899351682694
        """
        return self._species._propflow('S', self._mol, self.T, self.P, self._phase)

    # Gibbs free energy
    @property
    def G(self):
        """Gibbs free energy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.G # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.G
        -182.43024800100193
        """
        return self.H - self.S*self.T

    # Internal energy
    @property
    def U(self):
        """Internal energy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.U # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.U
        -7090.732710112934
        """
        return self.H - self.P*self.volnet

    # Helmholtz
    @property
    def A(self):
        """Helmholtz energy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.A # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.A
        -12371.980502651397
        """
        return self.U - self.T*self.S

    # Capacity flow rate
    @property
    def C(self):
        """Heat capacity flow rate as a function of T (kJ/K/hr).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s2.C
        243.41704115214097
        """
        return self._species._propflow('Cpm', self._mol, self.T, self.P, self._phase)

    # Material properties
    @property
    def Cp(self):
        """Specific heat capacity as a function of T (kJ/kg/K).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.Cp
        2.641906706110962
        """
        return self.Cpm*self.molnet/self.massnet

    @property
    def Cpm(self):
        """Molar heat capacity as a function of T (kJ/kmol/K).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.Cpm # (J/mol/K)
        121.70852057607048
        """
        return self.C/self.molnet

    @property
    def Vm(self):
        """Molar volume as a function of T and P (m^3/mol).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.Vm
        0.00012030150757118572
        """
        return self._species._prop('Vm', self._mol, self.T, self.P, self._phase)

    @property
    def rho(self):
        """Density as a function of T (kg/m^3).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.rho
        765.8830039638536
        """
        return self.massnet/self.volnet

    @property
    def rhom(self):
        """Molar density as a function of T and P (mol/m^3).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.rhom
        16.624895567634884
        """
        return self.molnet/self.volnet

    @property
    def nu(self):
        """Kinematic viscosity as a function of T and P (m^2/s).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.nu
        5.733020843964377e-06
        """
        return self.mu/self.rho

    @property
    def mu(self):
        """Hydraulic viscosity as a function of T and P (Pa*s).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.mu
        0.0010780252844121937
        """
        # Katti, P.K.; Chaudhri, M.M. (1964). "Viscosities of Binary Mixtures of Benzyl Acetate with Dioxane, Aniline, and m-Cresol". Journal of Chemical and Engineering Data. 9 (1964): 442–443.
        # molfrac = self.molfrac
        # props = self._species._props
        # mus = np.array(props('mu', self._mol, self.T, self.P, self._phase))
        # Vms = np.array(props('mu', self._mol, self.T, self.P, self._phase))
        # pos = np.where(molfrac != 0)
        # return np.exp((molfrac[pos]*np.log(mus[pos]*Vms[pos])).sum())/self.Vm
        return self._species._prop('mu', self._mol, self.T, self.P, self._phase)

    @property
    def k(self):
        """Thermal conductivity as a function of T and P (W/m/k).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.k
        0.16476716002011285
        """
        return self._species._prop('k', self._mol, self.T, self.P, self._phase)

    @property
    def alpha(self):
        """Thermal diffusivity as a function of T and P (m^2/s).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.alpha
        8.614274122585474e-08
        """
        return self._species._prop('alpha', self._mol, self.T, self.P, self._phase)

    @property
    def sigma(self):
        """Surface tension as a function of T (N/m).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.sigma
        0.02188440412824106
        """
        return self._species._prop('sigma', self._mol, self.T, self.P, self._phase)

    @property
    def Pr(self):
        """Prandtl number as a function of T and P (non-dimensional).

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.Pr
        15.923321696004477
        """
        return self._species._prop('Pr', self._mol, self.T, self.P, self._phase)

    @property
    def P_vapor(self):
        """Vapor pressure (Pa), not taking into account light species always in gas phase.
        
        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.P_vapor
        array([7872.1566667784855, 0 ])
        """
        mol = self.mol
        species = self._species
        indices = species._equilibrium_indices(mol>0)
        compounds = species._compounds
        N = len(indices)
        P_vapor = np.zeros_like(mol)
        if N==0: return P_vapor
        species = [compounds[i] for i in indices]
        mol = self.mol[indices]
        x = mol/mol.sum()
        T = self.T
        Psat = [s.VaporPressure(T) for s in species]
        self._gamma.species = species
        P_vapor[indices] = x * Psat * self._gamma(x, T)
        return P_vapor
        
    # Other properties
    @property
    def source(self):
        """Unit source."""
        return self._source

    @property
    def sink(self):
        """Unit sink."""
        return self._sink

    @property
    def nonzero_species(self):
        """Flow indices and species that have a non-zero flow rate.

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=2)
        >>> s1.nonzero_species
        [1], ['Water']
        """
        return nonzero_species(self._species, self.mol)
    
    def quantity(self, prop_ID):
        """Return a property as a Quantity object as described in the `pint package <https://pint.readthedocs.io/en/latest/>`__ 

        **Parameters**

             **prop_ID:** [str] name of the property (e.g. 'mol', 'H', 'k' ...)

        Example:

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.quantity('Cp')
        2.4337467143619698 J/K/g

        """
        attr = getattr(self, prop_ID)
        if isinstance(attr, np.ndarray):
            attr = np.array(attr.astype(float))
        return _Q(attr, self.units[prop_ID])

    # Derivative of a property and sensivity analysis
    def derivative(self, prop_ID, var_ID, index=None):
        """Return the derivative of property 'prop_ID' and variable 'var_ID'. If the property given by var_ID is an array, use index to specify which element.

         **Parameters**

             **prop_ID:** [str] Name of the property (e.g. 'rho', 'Cp', 'H', 'volnet' ...)

             **var_ID:** [str] Name of the variable (e.g 'T', 'P', 'molnet' ...)

             **index:** [array_like or None] Optional indices if the variable is an array

        **Example**

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Ethanol=2)
        >>> s1.derivative('rho', 'T')
        -0.8917376157800969

        """
        if index is not None:
            def getvar():
                return getattr(self, var_ID)[index]

            def setvar(xf):
                getattr(self, var_ID)[index] = xf
        else:
            def getvar():
                return getattr(self, var_ID)

            def setvar(xf):
                setattr(self, var_ID, xf)
        x0 = getvar()
        y0 = getattr(self, prop_ID)
        xf = x0 + 10**-6
        setvar(xf)
        yf = getattr(self, prop_ID)
        setvar(x0)
        return (yf-y0)/(xf-x0)

    # Specifications
    @staticmethod
    def like(stream, ID=''):
        """Create either a Stream or MixedStream object just like the given stream, depending on whether multiple phases are present."""
        s = stream
        if isinstance(stream, MS.MixedStream):
            out = MS.MixedStream(ID,
                                 species=stream._species,
                                 T=s.T, P=s.P)
            out._molarray[:] = stream._molarray
        else:
            out = Stream(ID, flow=s._mol, species=s._species,
                         phase=s._phase, T=s.T, P=s.P)
        return out
    
    def copylike(self, stream):
        """Copy flow rates, T, P, and phase of stream to self.

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream()
        >>> s2 = Stream(Ethanol=1, Water=2, T=400, phase='g')
        >>> s1.copylike(s2)
        >>> s1.show()
        Stream: s1
         phase: 'g', T: 400.00 K, P: 101325 Pa
         flow (kmol/hr): Ethanol  1
                         Water    2
        """
        if isinstance(stream, MS.MixedStream):
            self.enable_phases()
            self.copylike(stream)
        elif not (self._species is stream._species):
            raise ValueError('species must be the same to copy stream specifications')
        else:
            self._mol[:] = stream.mol
            self.P = stream.P
            self._phase = stream.phase
            self.T = stream.T
            
    def copyflow(self, stream, species=None, *, remove=False, exclude=False):
        """Copy flow rates of stream to self.
        
        **Parameters**
        
            **stream:** [Stream] Flow rates will be copied from here.
            
            **species:** iterable[str] Species IDs. Defaults to all species.
            
            **remove:** [bool] If True, copied species will be removed from `stream`.
            
            **exclude:** [bool] If True, exclude `species` when copying.
            
        .. Note::
            
            Species flow rates that are not copied will be set to zero.
        
        """
        assert self._species is stream._species, ('species must be the same to '
                                                  'copy stream specifications')
        if species is None:
            self._mol[:] = stream.mol
            if remove: stream._mol[:] = 0
        else:
            indices = self.indices(species)
            if exclude:
                self._mol[:] = stream.mol
                self._mol[indices] = 0
                if remove:
                    mol = stream._mol
                    if isinstance(stream, MS.MixedStream):
                        mol[:], mol[:, indices] = 0, mol[:, indices]
                    else:
                        mol[:], mol[indices] = 0, mol[indices]
            else:
                self._mol[:] = 0
                self._mol[indices] = stream.mol[indices]
                if remove: 
                    if isinstance(stream, MS.MixedStream):
                        stream._mol[phase_index[self.phase], indices] = 0
                    else:
                        stream._mol[indices] = 0

    def recieve_vent(self, stream, efficiency=1):
        assert self._species is stream._species, 'species must be the same to recieve vent'
        y = stream.P_vapor/self.P
        Y = y.sum()
        vent_mol = self._mol
        mol2vent = vent_mol.sum()*y*Y/(1-Y)*efficiency
        stream._mol[:] -= mol2vent
        vent_mol += mol2vent

    def empty(self):
        """Set flow rates to zero

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=1)
        >>> s1.empty()
        >>> s1.mol
        array([0, 0])
        """
        self._mol[:] = 0
    
    def bubble_T(self):
        """Bubble point at current composition and pressure.

        **Returns**

            **T:** [float] Bubble point temperature (T).

            **y:** [numpy ndarray] Vapor phase composition.
            
            **indices:** [list] Indices of species in equilibrium

        >>> from biosteam import *
        >>> stream = Stream(flow=(0.6, 0.4),
        ...                 species=Species('Ethanol', 'Water'))
        >>> stream.bubble_T()
        (352.2820850833474, array([0.703, 0.297]), [0, 1])

        """
        mol = self.mol
        # If just one specie in equilibrium, return Tsat
        indices = self._species._equilibrium_indices(mol>0)
        cmps = self._species._compounds
        N = len(indices)
        if N == 1:
            return (cmps[indices[0]].Tsat(self.P), np.array((1,)), indices)
        elif N == 0:
            raise EquilibriumError('no species available for phase equilibrium')
        mol = mol[indices]
        self._gamma.species = [cmps[i] for i in indices]
        # Solve and return bubble point
        return (*self._bubble_point.solve_Ty(mol/mol.sum(), self.P), indices)
    
    def bubble_P(self):
        """Bubble point at current composition and temperature.

        **Returns**

            **P:** [float] Bubble point pressure (Pa).

            **y:** [numpy ndarray] Vapor phase composition.
            
            **indices:** [list] Indices of species in equilibrium

        >>> from biosteam import *
        >>> s1 = Stream(flow=(0.703, 0.297), T=352.28,
                        species=Species('Ethanol', 'Water'))
        >>> s1.bubble_P()
        (103494.17209657285, array([0.757, 0.243]), [0, 1])

        """
        mol = self.mol
        # If just one specie in equilibrium, return Tsat
        indices = self._species._equilibrium_indices(mol>0)
        cmps = self._species._compounds
        N = len(indices)
        if N == 1:
            return (cmps[indices[0]].VaporPressure(self.T), np.array((1,)), indices)
        elif N == 0:
            raise EquilibriumError('no species available for phase equilibrium')
        mol = mol[indices]
        self._gamma.species = [cmps[i] for i in indices]
        # Solve and return bubble point
        return (*self._bubble_point.solve_Py(mol/mol.sum(), self.T), indices)

    def dew_T(self):
        """Dew point at current composition and pressure.
        
        **Returns**

            **T:** [float] Dew point temperature (K).

            **x:** [numpy array] Liquid phase composition.
            
            **indices:** [list] Indices of species in equilibrium

        >>> stream = Stream(flow=(0.5, 0.5),
        ...                 species=Species('Ethanol', 'Water'))
        >>> stream.dew_T()
        (357.45184742263075, array([0.151, 0.849]), [0, 1])
        
        """
        mol = self.mol
        # If just one specie in equilibrium, return Tsat
        indices = self._species._equilibrium_indices(mol>0)
        cmps = self._species._compounds
        N = len(indices)
        if N == 1:
            return (cmps[0].Tsat(self.P), np.array((1,)), indices)
        elif N == 0:
            raise EquilibriumError('no species available for phase equilibrium')
        mol = mol[indices]
        self._gamma.species = [cmps[i] for i in indices]
        # Solve and return dew point
        return (*self._dew_point.solve_Tx(mol/mol.sum(), self.P), indices)
    
    def dew_P(self):
        """Dew point at current composition and temperature.
        
        **Returns**

            **P:** [float] Dew point pressure (Pa).

            **x:** [numpy array] Liquid phase composition.
            
            **indices:** [list] Indices of species in equilibrium

        >>> from biosteam import *
        >>> stream = Stream(flow=(0.703, 0.297), T=352.28,
        ...                 species=Species('Ethanol', 'Water'))
        >>> stream.dew_P()
        (101328.47030327446, array([0.6, 0.4]), [0, 1])
        
        """
        mol = self.mol
        # If just one specie in equilibrium, return Tsat
        indices = self._species._equilibrium_indices(mol>0)
        cmps = self._species._compounds
        N = len(indices)
        if N == 1:
            return (cmps[0].VaporPressure(self.T), np.array((1,)), indices)
        elif N == 0:
            raise EquilibriumError('no species available for phase equilibrium')        
        mol = mol[indices]    
        self._gamma.species = [cmps[i] for i in indices]
        # Solve and return dew point
        return (*self._dew_point.solve_Px(mol/mol.sum(), self.T), indices)

    # Dimensionless number methods
    def Re(self, L, A=None):
        """Return Reynolds number.

        **Parameters**

             **L:** [float] Characteristic lenght (m)

             **A:** [float] Cross-sectional area (m^2). If A is None, assume flow through a cylindrical pipe.

        Example

        >>> from biosteam import *
        >>> Stream.species = Species('Ethanol', 'Water')
        >>> s1 = Stream(Water=200)
        >>> s1.Re(0.2, 0.031416)
        27639.956372923512
        """
        volnet = self.volnet/3600 # m3/s
        if A is None:  # Assume pipe
            return 4*volnet/(self.nu*np.pi*L)
        return volnet*L/(self.nu*A)

    # Class methods
    @classmethod
    def _helpdata(cls, prop):
        """Return information related to a property.

        **Parameters**

             **prop:** [str] name or description of a property

        Return

             **out:** tupple[str] where elements are:
                  * **ID:** [str] name of the property

                  * **Description:** [str] description of the property

                  * **Dependency:** [str] TP dependency

                  * **Units:** [str] units of measure

                  * **Datatype:** [str] type returned by property

        Example

        >>> Stream._helplist('k')
        ('k', 'thermal conductivity', 'TP', 'W/m/K', 'float')

        >>> Stream._helplist('viscosity')
        (('nu', 'kinematic viscosity', 'TP', 'm^2/s', 'float'),
         ('mu', 'hydraulic viscosity', 'TP', 'Pa*s', 'float'))

        >>> Stream._helplist('kmol/hr')
        (('mol', 'molar flow rates', '', 'kmol/hr', 'np.array'),
         ('molnet', 'net molar flow rate', '', 'kmol/hr', 'float'))
        """
        # Compare by both ID and description
        for index in (0, 1):
            # Search for exact matches
            for l in cls._prop_info:
                if prop == l[index]: return l
        
        out = []
        for index in (0, 1):    
            # If no matches found, search harder
            for l in cls._prop_info:
                if prop.lower() in l[index].lower(): out.append(l)
        return out

    @classmethod
    def help(cls, prop):
        """Print information related to a property.

        **Parameters**

             **prop:** [str] name or description of a property     

        Example

        >>> Stream.help('rho')
        rho: [float] Density as a function of T and P (kg/m^3).
        >>> Stream.help('density')
        rho: [float] Density as a function of T and P (kg/m^3).
        """
        data = cls._helpdata(prop)
        if data: _print_helpdata(data)
        else: print(f"No matching property '{prop}'.")

    @staticmethod
    def T_equilibrium(streams, T_guess=None, Q_in=0, approximate=True):
        """Bring all streams to temperature equilibrium.

        **Parameters**

            **streams:** [Stream] All streams
        
            **T_guess:** [float] Equilibrium temperature guess (K)

            **Q_in:** [float] Heat addition to streams (kJ/hr)     
            
            **approximate:** [bool] If True, approximate energy balance with constant heat capacity

        **Return**

             **T:** [float] New temperature (K)

        **Example**

        >>> from biosteam import *
        >>> Stream.species = Species('Water', 'Ethanol')
        >>> s1 = Stream(Water=2, T=300)
        >>> s2 = Stream(Ethanol=1, Water=2, T=340)
        >>> Stream.T_equilibrium([s1, s2])
        >>> (s1.T, s2.T)
        (325.8281231556553, 325.8281231556553)
        """
        sum_ = sum
        # Set up problem
        C = sum_([s.C for s in streams])
        H_ = sum_([s.H for s in streams]) + Q_in
        
        # Set initial temperature
        if not T_guess:
            T_guess = sum_([s.T for s in streams])/len(streams)
        for s in streams:
            s.T = T_guess
        
        # Find new T
        H = sum_([s.H for s in streams])
        T = T_guess + (H_ - H)/C
        
        # Update
        for s in streams:
            s.T = T_guess
        if approximate: return
        
        # Solve enthalpy by iteration
        it = 1
        while abs(T - T_guess) > 0.01:
            # Calculate
            H_ = sum_([s.H for s in streams])
            T = T_guess + (H_- H)/C
            
            # Check iteration
            if it > 40:
                raise SolverError(f"could not solve temperature")
            
            # Update
            T_guess = T
            for s in streams:
                s.T = T_guess
            it += 1

    @staticmethod
    def sum(s_sum, streams):
        """Mixes streams and sets resulting mass and energy in 's_sum'. Assumes the same pressure as streams[0], and no phase equilibrium. If 's_sum' is a Stream object, it assumes the phase does not change.

        **Parameters** 

             **s_sum:** [Stream] Container for the resulting mixture of streams

             **streams:** [tuple] Stream objects to be mixed

        .. Note:: Ignores contents of `s_sum` as if it was empty

        **Example**

        >>> from biosteam import *
        >>> Stream.species = Species('Water', 'Ethanol')
        >>> s1 = Stream(Water=2, T=300)
        >>> s2 = Stream(Ethanol=1, Water=2, T=340)
        >>> s_sum = Stream('s_sum')
        >>> Stream.sum(s_sum, [s1, s2])
        >>> s_sum.show()
        Stream: s_sum
         phase: 'l', T: 326.29 K, P: 101325 Pa
         flow (kmol/hr): Ethanol  1
                         Water    4
        """
        # Check if energy balance is required
        T_init = streams[0].T
        other_streams = streams[1:]
        energy_balance = any([T_init!=s.T for s in other_streams])
        if energy_balance: H = sum([s.H for s in streams])

        # Copy starting values
        s_sum.copylike(streams[0])

        # Mass balance
        inst = isinstance
        MStream = MS.MixedStream
        try:
            if inst(s_sum, MStream):
                # For MixedStream objects
                for s in other_streams:
                    if inst(s, MStream):
                        s_sum._mol[:] += s._mol
                    else:
                        s_sum._mol[phase_index[s._phase]] += s._mol
            else:
                # For Stream objects
                for s in other_streams:
                    s_sum._mol[:] += s.mol
        except Exception as Error:
            if not inst(s_sum, Stream):
                raise TypeError('s_sum must be a Stream object, not '
                                "'{type(s_sum).__name__}'")
            for s in streams:
                if not inst(s, Stream):
                    raise TypeError('streams must only contain Stream '
                                    f'objects, not {type(s).__name__} objects')
            raise Error

        # Energy Balance
        if energy_balance: s_sum.H = H

    # MixedStream compatibility
    def enable_phases(self):
        """Cast stream into a MixedStream object."""
        mol = self._mol
        self.__class__ = MS.MixedStream
        self._setflows(np.zeros((4, self._species._N)))
        self._mol[phase_index[self._phase]] = mol
        self._lL_split_cached = (None,)
        self._VLE = VLE(self)

    def disable_phases(self, phase):
        """Cast stream into a Stream object.
        
        **Parameters**
        
            **phase:** {'s', 'l', 'g'} desired phase of stream
            
        """
        self._phase = phase
            
    @property
    def VLE(self):
        """A callable VLE object for vapor-liquid equilibrium.

        **Parameters**
        
            **Specify two:**
                * **P:** [float] Operating pressure (Pa)
                * **Q:** [float] Energy input (kJ/hr)
                * **T:** [float] Operating temperature (K)
                * **V:** [float] Molar vapor fraction
                * **x:** [numpy array] Molar composition of liquid (for binary mixture)
                * **y:** [numpy array] Molar composition of vapor (for binary mixture)

            **Optional:**
            
                **species_IDs:** [tuple] IDs of species in equilibrium.
                     
                **LNK:** tuple[str] Light non-keys that remain as a vapor.
        
                **HNK:** tuple[str] Heavy non-keys that remain as a liquid.

        .. Note:
           LNK and HNK are not taken into account for equilibrium. Parameters not specified are None by default.

        """
        self.enable_phases()
        return self._VLE
        
    def LLE(self, species_IDs=(), split=None, lNK=(), LNK=(),
            solvents=(), solvent_split=(),
            P=None, T=None, Q=None):
        self.enable_phases()
        self.LLE(species_IDs, split, lNK, LNK,
                 solvents, solvent_split, P, T, Q)

    def _info_header(self):
        """Return stream information header."""
        # First line
        unit = self._source
        if unit is None:
            source = ''
        else:
            source = f'  from  {type(unit).__name__}-{unit}'
        unit = self._sink
        if unit is None:
            sink = ''
        else:
            sink = f'  to  {type(unit).__name__}-{unit}'
        if self.ID:
            return f"{type(self).__name__}: {self.ID}{source}{sink}"
        else:
            return f"{type(self).__name__}{source}{sink}"
        
    def _info_phaseTP(self, phases, T_units, P_units):
        T = _Q(self.T, self.units['T']).to(T_units).magnitude
        P = _Q(self.P, self.units['P']).to(P_units).magnitude
        return f" phase: '{phases}', T: {T:.5g} {T_units}, P: {P:.6g} {P_units}\n"

    # Representation
    def _info(self, T, P, flow, fraction):
        """Return string with all specifications."""
        units = self.units
        basic_info = self._info_header() + '\n'
        if hasattr(self, '_mol'):
            nonzero, species = self.nonzero_species
        else:
            return basic_info + f' link: {self._link}'
        T_units, P_units, flow_units, fraction = [(i if i is not None else j) for i, j in
                                                  zip((T, P, flow, fraction), self.display_units)]
        basic_info += self._info_phaseTP(self._phase, T_units, P_units)
        len_ = len(nonzero)
        if len_ == 0:
            return basic_info + ' flow: 0' 
        # Start of third line (flow rates)
        flow_dim = _Q(0, flow_units).dimensionality
        if fraction:
            if flow_dim == mol_flow_dim:
                flownet = _Q(self.molnet, units['molnet']).to(flow_units).magnitude
                flow = 'molfrac'
            elif flow_dim == mass_flow_dim:
                flownet = _Q(self.massnet, units['massnet']).to(flow_units).magnitude
                flow = 'massfrac'
            elif flow_dim == vol_flow_dim:
                flownet = _Q(self.volnet, units['volnet']).to(flow_units).magnitude
                flow = 'volfrac'
            else:
                raise DimensionError(f"dimensions for flow units must be in molar, mass or volumetric flow rates, not '{flow_dim}'")
            beginning = ' flow: '
            end = f'\n*{flownet:.3g} {flow_units} '
        else:
            beginning = f' flow ({flow_units}): '
            end = ''
            if flow_dim == mol_flow_dim:
                flow = 'mol'
            elif flow_dim == mass_flow_dim:
                flow = 'mass'
            elif flow_dim == vol_flow_dim:
                flow = 'vol'
            else:
                raise DimensionError(f"dimensions for flow units must be in molar, mass or volumetric flow rates, not '{flow_dim}'")
        # Remaining lines (all flow rates)
        new_line_spaces = len(beginning) * ' '
        if 'frac' in flow:
            flow = getattr(self, flow)[nonzero]
        else:
            flow = _Q(getattr(self, flow)[nonzero], units[flow]).to(flow_units).magnitude
        
        flowrates = ''
        lengths = [len(sp) for sp in species]
        maxlen = max(lengths) + 1
        for i in range(len_-1):
            spaces = ' ' * (maxlen - lengths[i])
            flowrates += species[i] + spaces + f' {flow[i]:.3g}\n' + new_line_spaces
        spaces = ' ' * (maxlen - lengths[len_-1])
        flowrates += species[len_-1] + spaces + f' {flow[len_-1]:.3g}'
        
        return (basic_info 
              + beginning
              + flowrates
              + end.replace('*', new_line_spaces + 'net' + (maxlen-4)*' ' + '  '))

    def show(self, T=None, P=None, flow=None, fraction=None):
        """Print all specifications."""
        print(self._info(T, P, flow, fraction))
    _ipython_display_ = show

    def _disconnect(self):
        outs = self._source and self._source._outs
        if outs: outs[outs.index(self)] = MissingStream
        ins = self._sink and self._sink._ins
        if ins: ins[ins.index(self)] = MissingStream
        self._source = self._sink = None

    def __str__(self):
        if self.ID: return self.ID
        else: return type(self).__name__

    def __repr__(self):
        if self.ID: return f'<{type(self).__name__}: {self.ID}>'
        else: return f'<{type(self).__name__}>'
        
        
from . import _mixed_stream as MS


# -*- coding: utf-8 -*-
"""
Created on Tue Jul  9 00:35:01 2019

@author: yoelr
"""
__all__ = ('NotImplementedMethod',)

class NotImplementedMethodType:
    __slots__ = ()
    def __new__(self): return NotImplementedMethod
    def __call__(self): pass
    def __bool__(self): return False
    def __repr__(self): return "NotImplementedMethod"

NotImplementedMethod = object.__new__(NotImplementedMethodType)


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 20 18:40:05 2019

@author: yoelr
"""

__all__ = []

from .vle import *
from .unifac import *
from .dew_point import *
from .bubble_point import *

from . import unifac
from . import vle
from . import dew_point
from . import bubble_point

__all__.extend(unifac.__all__)
__all__.extend(vle.__all__)
__all__.extend(dew_point.__all__)
__all__.extend(bubble_point.__all__)




# -*- coding: utf-8 -*-
"""
Created on Wed Jul 17 18:43:39 2019

@author: yoelr
"""
from . import Facility
import biosteam as bst

__all__ = ('ProcessWaterCenter',)

cost = bst.units.decorators.cost

@cost('Makeup water flow rate', 'Makeup water pump',
      CE=551, kW=20*0.7457, cost=6864, S=155564, n=0.8, BM=3.1)
@cost('Process water flow rate', 'Process water pump',
      CE=551, kW=75*0.7457, cost=15292, S=518924, n=0.8, BM=3.1)
@cost('Process water flow rate', 'Tank',
      CE=522, cost=250e3, S=451555, n=0.7, BM=1.7)
class ProcessWaterCenter(Facility):
    """
    **ins**
    
        [0] Recycle water
        
        [1] Make-up water
        
        [2] Boiler water
        
        [3] Cooling tower water
        
    **outs**
    
        [0] Process water
        
        [1] Boiler water
        
        [2] Cooling tower water
    
    """
    _N_ins = 4
    _N_outs = 3
    boiler_blowdown = 0.002
    RO_rejection = 0.25
    CT_evaporation = 0.01
    CT_blowdown = 0.001
    _units = {'Makeup water flow rate': 'kg/hr',
              'Process water flow rate': 'kg/hr'}
    def __init__(self, ID='', ins=None, outs=(), *, BT, CT):
        Facility.__init__(self, ID, ins, outs)
        self.BT = BT
        self.CT = CT
        
    def _run(self):
        s_recycle, s_makeup, s_BT, s_CT = self._ins
        s_process, s_BT_, s_CT_ = self.outs
        s_BT.link = s_BT_
        s_CT.link = s_CT_
        process_water = s_process.molnet
        recycle_water = s_recycle.molnet
        wi = s_BT.index('7732-18-5')
        s_BT.mol[wi] = BT_water = self.BT.total_steam * self.boiler_blowdown * 1/(1-self.RO_rejection)
        wi = s_CT.index('7732-18-5')
        s_CT.mol[wi] = CT_water = self.CT.cooling_water * (self.CT_evaporation + self.CT_blowdown)
        process_loss = process_water - recycle_water
        makeup_water = process_loss if process_loss > 0 else 0
        wi = s_makeup.index('7732-18-5')
        s_makeup.mol[wi] = makeup_water
        Design = self._Design
        Design['Process water flow rate'] = (process_water + BT_water + CT_water)*18.015
        Design['Makeup water flow rate'] = (makeup_water + BT_water + CT_water)*18.015
        
        


# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 14:40:28 2018

@author: yoelr
"""
import re
import numpy as np
import pandas as pd
from graphviz import Digraph
from ._exceptions import DesignWarning, _try_method
from ._flowsheet import find, save_digraph
from ._graphics import Graphics, default_graphics
from ._stream import Stream
from ._heat_utility import HeatUtility
from ._utils import Ins, Outs, MissingStream, NotImplementedMethod
from ._power_utility import PowerUtility
from warnings import warn
import biosteam as bst

__all__ = ('Unit',)

# %% Unit neighbors

def _add_upstream_neighbors(unit, set):
    """Add upsteam neighboring units to set."""
    for s in unit._ins:
        u_source = s._source
        if u_source: set.add(u_source)

def _add_downstream_neighbors(unit, set):
    """Add downstream neighboring units to set."""
    for s in unit._outs:
        u_sink = s._sink
        if u_sink: set.add(u_sink)


# %% Bounds checking

def _warning(source, msg, category=Warning):
        """Return a Warning object with source description."""
        if isinstance(source, str):
            msg = f'@{source}: ' + msg
        elif source:
            msg = f'@{type(source).__name__} {str(source)}: ' + msg
        return category(msg)
            
def _lb_warning(key, value, units, lb, stacklevel, source):
    units = ' ' + units if units else ''
    try:
        msg = f"{key} ({value:.4g}{units}) is out of bounds (minimum {lb:.4g}{units})."
    except:  # Handle format errors
        msg = f"{key} ({value:.4g}{units}) is out of bounds (minimum {lb}{units})."
    
    warn(_warning(source, msg, DesignWarning), stacklevel=stacklevel)
    
def _ub_warning(key, value, units, ub, stacklevel, source):
    units = ' ' + units if units else ''
    try:
        msg = f"{key} ({value:.4g}{units}) is out of bounds (maximum {ub:.4g}{units})."
    except:  # Handle format errors
        msg = f"{key} ({value:.4g}{units}) is out of bounds (maximum {ub}{units})."
    
    warn(_warning(source, msg, DesignWarning), stacklevel=stacklevel)

# %% Metaclass for Unit operations
    
class unit(type):
    """Unit metaclass for keeping track for Unit lines and graphics."""
    def __new__(mcl, name, bases, dct, isabstract=False, unitdoc=True):
        # Make new Unit class
        cls = type.__new__(mcl, name, bases, dct)
        
        try: Unit
        except NameError: return cls
        
        # Set line
        # default_line constitutes a new Unit class
        line = cls.line
        if line in ('Unit', 'Mixer', 'Static', 'Splitter', 'Solids separator', 'Facility'):
            line = cls.__name__.replace('_', ' ')
            # Set new graphics object for new line
            if '_graphics' not in dct: cls._graphics = Graphics.box(cls._N_ins, cls._N_outs)
        elif 'line' in dct and '_graphics' not in dct:
            # Set new graphics for specified line
            cls._graphics = Graphics.box(cls._N_ins, cls._N_outs)
        
        cls.line = re.sub(r"\B([A-Z])", r" \1", line).capitalize()
        
        if unitdoc and cls.__doc__:
            cls.__doc__ = cls.__doc__.replace('**Parameters**', 
   "**Parameters**"
"\n"    
"\n        **ID:** [str] Unique identification. If set as '', a default ID will be chosen."
"\n"
"\n        **outs:** tuple[str or Stream] Output streams or IDs to initialize output streams. If None, leave streams missing. If empty, default IDs will be given."
"\n"
"\n        **ins:** tuple[str or Stream] Input streams or IDs to initialize input streams. If None, leave streams missing. If empty, default IDs will be given.")
        
        if isabstract: return cls
        
        if not hasattr(cls, '_run'):
            raise NotImplementedError("'Unit' subclass must have a '_run' method unless the 'isabstract' keyword argument is True")
        
        return cls


# %% Unit Operation

class Unit(metaclass=unit):
    """Abstract parent class for Unit objects. Child objects must contain _run, _design and _cost methods to estimate stream outputs of a Unit and find design and cost information.  

    **Parameters**

        **ID:** [str] A unique identification. If ID is an empty string (i.e. '' ), a default ID will be chosen. If ID is None, unit will not be registered in flowsheet.

        **ins:** tuple[str or Stream] Input streams or IDs to initialize input streams. If None, leave streams missing. If empty, default IDs will be given.

        **outs:** tuple[str or Stream] Output streams or IDs to initialize output streams. If None, leave streams missing. If empty, default IDs will be given.
                

    **Class Definitions** 
    
        **line** = [Defaults to the class name of the first child class]: [str] Name denoting the type of Unit class    
        
        **BM** = None: [float] Bare module factor (installation factor)
        
        **_run()**
            Run simulation and update output streams.

        **_design()**
            Add design requirements to "_Design" dictionary attribute.

        **_cost()**
            Add itemized purchse costs to results "_Cost" dictionary attribute.
            
        .. Note::
           
           The `_run` method is called during recycle loop convergence. The `_design` and `_cost` methods are called in the given order for generating results.
        
        **_units** = {}: [dict] Default units for results Operation and Design
        
        **_N_ins** = 1: [int] Expected number of input streams

        **_N_outs** = 2: [int] Expected number of output streams

        **_N_heat_utilities** = 0: [int] Number of heat utilities  

        **_has_power_utility** = False: [bool] If True, a PowerUtility object is created for every instance.
        
        **_has_cost** = True: [bool] Should be True if it has any associated cost
    
        **_graphics** = <Graphics>: [biosteam Graphics] Settings for diagram representation.

    **ins**
        
        list of input streams
        
    **outs**
    
        list of output streams
    
    **Examples**
    
        :doc:`Creating a Unit`
        
        :doc:`Using -pipe- notation`
        
        :doc:`Inheriting from Unit`
        
        :doc:`Unit decorators`
    
    """ 
    ### Abstract Attributes ###
    
    #: [float] Bare module factor (installation factor).
    BM = None
    
    # [dict] Default units for construction
    _units = {}
    
    # [int] Expected number of input streams
    _N_ins = 1  
    
    # [int] Expected number of output streams
    _N_outs = 2  
    
    # [int] number of heat utilities
    _N_heat_utilities = 0
    
    # [PowerUtility] A PowerUtility object, if required
    _power_utility = None
    
    # [bool] If True, a PowerUtility object is created for every instance.
    _has_power_utility = False 
    
    # [biosteam Graphics] a Graphics object for diagram representation.
    _graphics = default_graphics
    
    # [str] The general type of unit, regardless of class
    line = 'Unit'

    ### Other defaults ###

    # [list] Default ID starting letter and number
    _default_ID = ['U', 1]
    
    # Default ID
    _ID = None 
    
    #: [list] HeatUtility objects associated to unit
    _heat_utilities = ()
        
    ### Initialize ###
    
    def __init__(self, ID='', ins=None, outs=()):
        self._init_ins(ins)
        self._init_outs(outs)
        self._init_results()
        self._init_heat_utils()
        self._init_power_util()
        self.ID = ID

    def _init_ins(self, ins):
        """Initialize input streams."""
        if ins is None:
            self._ins = Ins(self, (MissingStream for i in range(self._N_ins)))
        elif isinstance(ins, Stream):
            self._ins = Ins(self, (ins,))
        elif isinstance(ins, str):
            self._ins = Ins(self, (Stream(ins),))
        elif not ins:
            self._ins = Ins(self, (Stream('') for i in range(self._N_ins)))
        else:
            self._ins = Ins(self, (i if isinstance(i, Stream) else Stream(i) for i in ins))
    
    def _init_outs(self, outs):
        """Initialize output streams."""
        if outs is None:
            self._outs = Outs(self, (MissingStream for i in range(self._N_outs)))
        elif not outs:
            self._outs = Outs(self, (Stream('') for i in range(self._N_outs)))
        elif isinstance(outs, Stream):
            self._outs = Outs(self, (outs,))
        elif isinstance(outs, str):
            self._outs = Outs(self, (Stream(outs),))
        else:
            self._outs = Outs(self, (i if isinstance(i, Stream) else Stream(i) for i in outs))        
    
    def _init_results(self):
        """Initialize attributes to store results."""
        # [dict] Updated in `_cost` method
        self._Cost = {}
        
        # [dict] Updated in `_design` method
        self._Design = {}
        
        # [dict] Greenhouse gas emissions
        self._GHGs = {}
    
    def _init_heat_utils(self):
        """Initialize heat utilities."""
        if self._N_heat_utilities: 
            self._heat_utilities = [HeatUtility() for i in
                                    range(self._N_heat_utilities)]
        
    def _init_power_util(self):
        """Initialize power utility."""
        if self._has_power_utility:
            self._power_utility = PowerUtility()         
    
    # Forward pipping
    def __sub__(self, other):
        """Source streams."""
        if isinstance(other, Unit):
            other._ins[:] = self._outs
            return other
        elif type(other) is int:
            return self.outs[other]
        elif isinstance(other, Stream):
            self._outs[:] = (other,)
            return self
        elif isinstance(other, (tuple, list, np.ndarray)):
            if isinstance(other[0], int):
                return [self.outs[i] for i in other]
            else:
                self._outs[:] = other
                return self
        else:
            return other.__rsub__(self)

    def __rsub__(self, other):
        """Sink streams."""
        if type(other) is int:
            return self._ins[other]
        elif isinstance(other, Stream):
            self._ins[:] = (other,)
            return self
        elif isinstance(other, (tuple, list, np.ndarray)):
            if all(isinstance(i, int) for i in other):
                return [self._ins[i] for i in other]
            else:
                self._ins[:] = other
                return self

    # Backwards pipping
    __pow__ = __sub__
    __rpow__ = __rsub__
    
    # Abstract methods
    _design   = NotImplementedMethod
    _cost     = NotImplementedMethod
    
    # Summary
    def _summary(self):
        """Calculate all results from unit run."""
        self._design()
        self._cost()
    
    @property
    def purchase_cost(self):
        """Total purchase cost (USD)."""
        return sum(self._Cost.values())
    
    @property
    def installation_cost(self):
        """Installation cost (USD)."""
        return self.BM * sum(self._Cost.values())
    
    @property
    def utility_cost(self):
        """Total utility cost (USD/hr)."""
        if self._power_utility:
            return (sum([i.cost for i in self._heat_utilities])
                                 + self._power_utility.cost)
        else:
            return sum([i.cost for i in self._heat_utilities])

    def simulate(self):
        """Run rigourous simulation and determine all design requirements."""
        _try_method(self._run)
        _try_method(self._summary)

    def results(self, with_units=True):
        """Return key results from simulation as a DataFrame if `with_units` is True or as a Series otherwise."""
        ID = self.ID
        keys = []; addkey = keys.append
        vals = []; addval = vals.append
        if with_units:
            if self._power_utility:
                i = self._power_utility
                addkey(('Power', 'Rate'))
                addkey(('Power', 'Cost'))
                addval(('kW', i.rate))
                addval(('USD/hr', i.cost))
            if self._heat_utilities:
                for i in self._heat_utilities:
                    addkey((i.ID, 'Duty'))
                    addkey((i.ID, 'Flow'))
                    addkey((i.ID, 'Cost'))
                    addval(('kJ/hr', i.duty))
                    addval(('kmol/hr', i.flow))
                    addval(('USD/hr', i.cost))
            units = self._units
            Cost = self._Cost
            for ki, vi in self._Design.items():
                addkey(('Design', ki))
                addval((units.get(ki, ''), vi))
            for ki, vi in Cost.items():
                addkey(('Cost', ki))
                addval(('USD', vi))
            addkey(('Purchase cost', ''))
            addval(('USD', self.purchase_cost))
            addkey(('Utility cost', ''))
            addval(('USD/hr', self.utility_cost))
            if self._GHGs:
                a, b = self._totalGHG
                GHG_units =  self._GHG_units
                for ko, vo in self._GHGs.items():
                    for ki, vi in vo.items():
                        addkey((ko, ki))
                        addval((GHG_units.get(ko, ''), vi))
                a_key, b_key = GHG_units.keys()
                a_unit, b_unit = GHG_units.values()
                addkey(('Total ' + a_key, ''))
                addval((a_unit, a))
                addkey(('Total ' + b_key, ''))
                addval((b_unit, b))
            df = pd.DataFrame(vals,
                              pd.MultiIndex.from_tuples(keys),
                              ('Units', ID))
            df.columns.name = self.line
            return df
        else:
            if self._power_utility:
                i = self._power_utility
                addkey(('Power', 'Rate'))
                addkey(('Power', 'Cost'))
                addval(i.rate)
                addval(i.cost)
            if self._heat_utilities:
                for i in self._heat_utilities:
                    addkey((i.ID, 'Duty'))
                    addkey((i.ID, 'Flow'))
                    addkey((i.ID, 'Cost'))
                    addval(i.duty)
                    addval(i.flow)
                    addval(i.cost)
            for ki, vi in self._Design.items():
                addkey(('Design', ki))
                addval(vi)
            for ki, vi in self._Cost.items():
                addkey(('Cost', ki))
                addval(vi)    
            if self._GHGs:
                GHG_units =  self._GHG_units
                for ko, vo in self._GHGs.items():
                    for ki, vi in vo.items():
                        addkey((ko, ki))
                        addval(vi)
                a, b = self._totalGHG
                a_key, b_key = GHG_units.keys()
                addkey(('Total ' + a_key, ''))
                addval(a)
                addkey(('Total ' + b_key, ''))
                addval(b)
            addkey(('Purchase cost', ''))
            addval(self.purchase_cost)
            addkey(('Utility cost', ''))
            addval(self.utility_cost)
            series = pd.Series(vals, pd.MultiIndex.from_tuples(keys))
            series.name = ID
            return series

    # def __getattr__(self, name):
    #     if name is '_results':
    #         warn(DeprecationWarning("'_results' dictionary will be deprecated in favor of attributes '_Design', '_Cost', and '_GHGs'"))
    #         self._results = results = {'Design': self._Design,
    #                                    'Cost': self._Cost,
    #                                    'GHG': self._GHGs}
    #         return results
    #     elif name in ('_Cost', '_Design') and hasattr(self, '_results'):
    #         warn(DeprecationWarning("'_results' dictionary will be deprecated in favor of attributes '_Design', '_Cost', and '_GHGs'"))
    #         self._Cost = self._results['Cost']
    #         self._Design = self._results['Design']
    #         self._GHGs = self._results['GHG']
    #     else:
    #         raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def _checkbounds(self, key, value, units, bounds):
        """Issue a warning if value is out of bounds.
        
        **Parameters**
        
            **key:** [str] Name of value.
            
            **value:** [float]
            
            **units:** [str] Units of value
            
            **bounds:** iterable[float, float] Upper and lower bounds.
            
        """
        # Warn when value is out of bounds
        lb, ub = bounds
        if not lb<=value and ub>=value:
            units = ' ' + units if units else ''
            try:
                msg = f"{key} ({value:.4g}{units}) is out of bounds ({lb:.4g} to {ub:.4g}{units})."
            except:  # Handle format errors
                msg = f"{key} ({value:.4g}{units}) is out of bounds ({lb} to {ub}{units})."
            warn(_warning(self, msg, DesignWarning), stacklevel=3)
    def _lb_warning(self, key, value, units, lb):
        """Warn that value is below lower bound.
        
         **Parameters**
        
            **key:** [str] Name of value.
            
            **value:** [float]
            
            **units:** [str] Units of value.
            
            **lb:** [float] lower bound.
        
        """
        _lb_warning(key, value, units, lb, 4, self)

    @property
    def ID(self):
        """Unique Identification (str). If set as '', it will choose a default ID."""
        return self._ID

    @ID.setter
    def ID(self, ID):
        if ID == '':
            # Select a default ID if requested
            letter, number = self._default_ID
            self._default_ID[1] += 1
            ID = letter + str(number)
            self._ID = ID
            find.unit[ID] = self
        elif ID and ID != self._ID:
            ID = ID.replace(' ', '_')
            ID_words = ID.split('_')
            if not all(word.isalnum() for word in ID_words):
                raise ValueError('ID cannot have any special characters')
            self._ID = ID
            find.unit[ID] = self

    # Input and output streams
    @property
    def ins(self):
        # list of input streams
        return self._ins    
    @property
    def outs(self):
        # list of output streams
        return self._outs

    @property
    def _downstream_units(self):
        """Return set of all units downstreasm."""
        downstream_units = set()
        outer_periphery = set()
        _add_downstream = _add_downstream_neighbors
        _add_downstream(self, outer_periphery)
        inner_periphery = None
        old_length = -1
        new_length = 0
        while new_length != old_length:
            old_length = new_length
            inner_periphery = outer_periphery
            downstream_units.update(inner_periphery)
            outer_periphery = set()
            for unit in inner_periphery:
                _add_downstream(unit, outer_periphery)
            new_length = len(downstream_units)
        return downstream_units

    def _neighborhood(self, radius=1):
        """Return all neighboring units within given radius.
        
        **Parameters**
        
            **radius:**[int] Maxium number streams between neighbors.
        
        """
        radius -= 1
        neighborhood = set()
        if radius < 0: return neighborhood
        _add_upstream_neighbors(self, neighborhood)
        _add_downstream_neighbors(self, neighborhood)
        direct_neighborhood = neighborhood
        for i in range(radius):
            neighbors = set()
            for neighbor in direct_neighborhood:
                _add_upstream_neighbors(neighbor, neighbors)
                _add_downstream_neighbors(neighbor, neighbors)
            if neighbors == direct_neighborhood: break
            direct_neighborhood = neighbors
            neighborhood.update(direct_neighborhood)
        
        return neighborhood

    def diagram(self, radius=0, file=None, format='png'):
        """Display a `Graphviz <https://pypi.org/project/graphviz/>`__ diagram of the unit and all neighboring units within given radius.
        
        **Parameters**
        
            **radius:** [int] Maxium number streams between neighbors.
        
            **file:** Must be one of the following:
                * [str] File name to save diagram.
                * [None] Display diagram in console.
        
            **format:** Format of file.
        
        """
        if radius > 0:
            neighborhood = self._neighborhood(radius)
            neighborhood.add(self)
            sys = bst.System('', neighborhood)
            return sys.diagram('thorough', file, format)
        
        graphics = self._graphics

        # Make a Digraph handle
        f = Digraph(name='unit', filename='unit', format='svg')
        f.attr('graph', ratio='0.5', splines='normal', outputorder='edgesfirst',
               nodesep='1.1', ranksep='0.8', maxiter='1000')  # Specifications
        f.attr(rankdir='LR')  # Left to right

        # If many streams, keep streams close
        if (len(self.ins) >= 3) or (len(self.outs) >= 3):
            f.attr('graph', nodesep='0.4')

        # Initialize node arguments based on unit and make node
        type_ = graphics.node_function(self) or self.line
        name = self.ID + '\n' + type_
        f.attr('node', **self._graphics.node)
        f.node(name)

        # Set stream node attributes
        f.attr('node', shape='rarrow', fillcolor='#79dae8',
               style='filled', orientation='0', width='0.6',
               height='0.6', color='black', peripheries='1')

        # Make nodes and edges for input streams
        di = 0  # Destination position of stream
        for stream in self.ins:
            if not stream: continue
            f.node(stream.ID)  
            edge_in = self._graphics.edge_in
            if di >= len(edge_in): di = 0
            f.attr('edge', arrowtail='none', arrowhead='none',
                   tailport='e', **edge_in[di])
            f.edge(stream.ID, name)
            di += 1

        # Make nodes and edges for output streams
        oi = 0  # Origin position of stream
        for stream in self.outs:
            if not stream: continue
            f.node(stream.ID) 
            edge_out = self._graphics.edge_out  
            if oi >= len(edge_out): oi = 0
            f.attr('edge', arrowtail='none', arrowhead='none',
                   headport='w', **edge_out[oi])
            f.edge(name, stream.ID)
            oi += 1
        save_digraph(f, file, format)
    
    ### Net input and output flows ###
    
    # Molar flow rates
    @property
    def _mol_in(self):
        """Molar flows going in (kmol/hr)."""
        return sum(s.mol for s in self._ins)

    @property
    def _mol_out(self):
        """Molar flows going out (kmol/hr)."""
        return sum(s.mol for s in self._outs)

    @property
    def _molfrac_in(self):
        """Molar fractions going in (kmol/hr)."""
        return self._mol_in/self._molnet_in

    @property
    def _molfrac_out(self):
        """Molar fractions going in."""
        return self._mol_out/self._molnet_out

    @property
    def _molnet_in(self):
        """Net molar flow going in (kmol/hr)."""
        return sum(s.molnet for s in self._ins)

    @property
    def _molnet_out(self):
        """Net molar flow going out (kmol/hr)."""
        return sum(s.molnet for s in self._outs)

    # Mass flow rates
    @property
    def _mass_in(self):
        """Mass flows going in (kg/hr)."""
        return sum(s.mass for s in self._ins)

    @property
    def _mass_out(self):
        """Mass flows going out (kg/hr)."""
        return sum(s.mass for s in self.outs)

    @property
    def _massfrac_in(self):
        """Mass fractions going in."""
        return self._mass_in/self._massnet_in

    @property
    def _massfrac_out(self):
        """Mass fractions going out."""
        return self._mass_out/self._massnet_out

    @property
    def _massnet_in(self):
        """Net mass flow going in (kg/hr)."""
        return sum(s.massnet for s in self._ins)

    @property
    def _massnet_out(self):
        """Net mass flow going out (kg/hr)."""
        return sum(s.massnet for s in self.outs)

    # Volumetric flow rates
    @property
    def _vol_in(self):
        """Volumetric flows going in (m3/hr)."""
        return sum(s.vol for s in self._ins)

    @property
    def _volnet_in(self):
        """Net volumetric flow going in (m3/hr)."""
        return sum(self._vol_in)

    @property
    def _volfrac_in(self):
        """Volumetric fractions going in."""
        return self._vol_in/self._volnet_in

    @property
    def _vol_out(self):
        """Volumetric flows going out (m3/hr)."""
        return sum([s.vol for s in self.outs])

    @property
    def _volnet_out(self):
        """Net volumetric flow going out (m3/hr)."""
        return sum(self._vol_out)

    @property
    def _volfrac_out(self):
        """Volumetric fractions going out."""
        return self._vol_out/self._volnet_out

    # Enthalpy flow rates
    @property
    def _H_in(self):
        """Enthalpy flow going in (kJ/hr)."""
        return sum([s.H for s in self._ins])

    @property
    def _H_out(self):
        """Enthalpy flow going out (kJ/hr)."""
        return sum([s.H for s in self._outs])

    @property
    def _Hf_in(self):
        """Enthalpy of formation flow going in (kJ/hr)."""
        return sum([s.Hf for s in self._ins])

    @property
    def _Hf_out(self):
        """Enthalpy of formation flow going out (kJ/hr)."""
        return sum([s.Hf for s in self._outs])

    @property
    def _Hnet(self):
        """Net enthalpy flow (including enthalpies of formation)."""
        return self._H_out - self._H_in + self._Hf_out - self._Hf_in
    
    # Representation
    def _info(self, T, P, flow, fraction):
        """Information on unit."""
        if self.ID:
            info = f'{type(self).__name__}: {self.ID}\n'
        else:
            info = f'{type(self).__name__}\n'
        info+= f'ins...\n'
        i = 0
        for stream in self._ins:
            if not stream:
                info += f'[{i}] {stream}\n'
                i += 1
                continue
            stream_info = stream._info(T, P, flow, fraction)
            unit = stream._source
            index = stream_info.index('\n')
            source_info = f'  from  {type(unit).__name__}-{unit}\n' if unit else '\n'
            info += f'[{i}] {stream.ID}' + source_info + stream_info[index+1:] + '\n'
            i += 1
        info += f'outs...\n'
        i = 0
        for stream in self._outs:
            if not stream:
                info += f'[{i}] {stream}\n'
                i += 1
                continue
            stream_info = stream._info(T, P, flow, fraction)
            unit = stream._sink
            index = stream_info.index('\n')
            sink_info = f'  to  {type(unit).__name__}-{unit}\n' if unit else '\n'
            info += f'[{i}] {stream.ID}' + sink_info + stream_info[index+1:] + '\n'
            i += 1
        info = info.replace('\n ', '\n    ')
        return info[:-1]

    def show(self, T=None, P=None, flow=None, fraction=None):
        """Prints information on unit."""
        print(self._info(T, P, flow, fraction))
    
    def _ipython_display_(self):
        try: self.diagram()
        except: pass
        self.show()
    
    def _disconnect(self):
        for i in self._ins:
            if i: 
                if i._source: i._sink = None
                else: object.__delattr__(find.stream, i._ID)
        for i in self._outs:
            if i:
                if i._sink: i._source = None
                else: object.__delattr__(find.stream, i._ID)
        self._outs.clear()
        self._ins.clear()
    
    def __str__(self):
        if self.ID:
            return self.ID
        else:
            return type(self).__name__

    def __repr__(self):
        if self.ID:
            return f'<{type(self).__name__}: {self.ID}>'
        else:
            return f'<{type(self).__name__}>'

_Unit_is_done = True



# -*- coding: utf-8 -*-
"""
Created on Mon May 20 22:04:02 2019

@author: yoelr
"""
from .. import Unit, Stream
from numpy import asarray
from pandas import Series

__all__ = ('run_split', 'run_split_with_mixing')


def run_split(self):
    """Splitter mass and energy balance function based on one input stream."""
    top, bot = self.outs
    feed = self._ins[0]
    net_mol = feed.mol
    top._mol[:] = net_mol * self._split
    bot._mol[:] = net_mol - top._mol
    bot.T = top.T = feed.T
    bot.P = top.P = feed.P

def run_split_with_mixing(self):
    """Splitter mass and energy balance function with mixing all input streams."""
    top, bot = self._outs
    ins = self._ins
    if len(ins) > 1: Stream.sum(top, ins)
    else: top.copylike(ins[0])
    bot.copylike(top)
    top._mol[:] *= self._split
    bot._mol[:] -= top._mol


class Splitter(Unit):
    """Create a splitter that separates mixed streams based on splits.

    **Parameters**

        **split:** Should be one of the following
            * [float] The fraction of net feed in the 0th output stream
            * [array_like] Componentwise split of feed to 0th output stream
            
        **order:** Iterable[str] Species order of split
    
    **ins**
    
        [0] Feed stream
        
    **outs**
    
        [0] Split stream
        
        [1] Remainder stream
    
    **Examples**
    
        Create a Splitter object with an ID, a feed stream, two output streams, and an overall split:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, units
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'), split=0.1)
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] top
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  1
           [1] bot
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  9
          
        Create a Splitter object, but this time with a componentwise split:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, units
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
           ...                     split=(0.1, 0.99))
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] top
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  9.9
           [1] bot
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  0.1
                               
        Create a Splitter object using componentwise split, but this time specify the order:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, units
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
           ...                     split=(0.99, 0.01), order=('Ethanol', 'Water'))
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] top
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  9.9
           [1] bot
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  0.1
    
    """
    _N_outs = 2
    
    @property
    def split(self):
        """[array] Componentwise split of feed to 0th output stream."""
        return self._split_series
    
    def __init__(self, ID='', ins=None, outs=(), *, split, order=None):
        Unit.__init__(self, ID, ins, outs)
        self._split = split = Stream.species.array(order, split) if order else asarray(split)
        self._split_series =  Series(split, Stream.species.IDs)

    def _run(self):
        top, bot = self._outs
        feed = self._ins[0]
        net_mol = feed.mol
        bot.T = top.T = feed.T
        bot.P = top.P = feed.P
        bot._phase = top._phase = feed._phase
        top._mol[:] = net_mol * self._split
        bot._mol[:] = net_mol - top._mol


class InvSplitter(Unit):
    """Create a splitter that sets the input stream based on output streams. Must have only one input stream. The output streams will become the same temperature, pressure and phase as the input.
    """
    line = 'Splitter'
    _graphics = Splitter._graphics
    def _run(self):
        feed = self.ins[0]
        feed._mol[:] = self._mol_out
        T = feed.T
        P = feed.P
        phase = feed.phase
        for out in self._outs:
            out.T = T
            out.P = P
            out.phase = phase 
    


        
    


# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 14:25:34 2018

@author: yoelr
"""
from ._exceptions import DimensionError
from ._utils import DisplayUnits
from ._species import Species
from ._stream import Stream, mol_flow_dim, mass_flow_dim, vol_flow_dim
from . import _Q


__all__ = ('HeatUtility',)

# Costs from Table 17.1 in Warren D.  Seider et al.-Product and Process Design Principles_ Synthesis, Analysis and Evaluation-Wiley (2016)
# ^This table was made using data from Busche, 1995
# Entry temperature conditions of coolants taken from Table 12.1 in Warren, 2016

# %% Default Heat Transfer species

_Water = Species('Water')


# %% Utility classes


class HeatUtility:
    """Create an HeatUtility object that can choose a utility stream and calculate utility requirements. It can calculate required flow rate, temperature, or phase change of utility. Calculations assume counter current flow rate.
    
    **Parameters**
    
        **efficiency:** [float] Fraction of heat transfered after accounting for heat loss.
    
    **Class Attributes**

        **cooling_agents:** dict[str: UtilityAgent] All cooling utilities available
        
        **heating_agents:** dict[str: UtilityAgent] All heating utilities available
    
    **Examples**
    
        Create a heat utility:
            
        .. code-block:: python
        
           >>> hu = HeatUtility()
           >>> hu
           HeatUtility: None
            duty: 0
            flow: 0
            cost: 0
        
        Calculate utility requirement by calling it with a duty (kJ/hr), and enter and exit temperature (K):
            
        .. code-block:: python
        
           >>> hu(1000, 300, 350)
           >>> hu
           HeatUtility: Low pressure steam
            duty: 1.11e+03 kJ/hr
            flow: 0.0284 kmol/hr,
            cost: 0.00674 USD/hr
       
        All results are accessible:
            
        .. code-block:: python
        
           >>> hu.ID, hu.duty, hu.flow, hu.cost
           ('Low pressure steam',
            1111.111111111111,
            0.028351477551759364,
            0.006741981361808377)
           
    """
    __slots__ = ('_fresh', '_used', 'ID', 'duty',
                 'flow', 'cost', 'efficiency',
                 '_args')
    dT = 5  #: [float] Pinch temperature difference
    
    #: Units of measure for results dictionary
    _units = dict(duty='kJ/hr', flow='kmol/hr', cost='USD/hr')
    
    #: [DisplayUnits] Units of measure for IPython display
    display_units = DisplayUnits(**_units)

    class UtilityAgent:
        """Create a UtilityAgent object that defines a utility option.
        
            **Parameters**
        
                species: [Species] 
                
                molfrac: [tuple] Molar fraction of species
                
                phase: [str] {'g', 'l'}
                
                Hvap: [float] Latent heat of vaporization (kJ/kmol)
                
                T_limit: [float] Temperature limit (K)
                
                price_kJ: [float] Price (USD/kJ)
                
                price_kmol: [float] Price (USD/kmol)
                
                efficiency: [float] Heat transfer efficiency (between 0 to 1)
                
        """
        __slots__ = ('species', 'molfrac', 'T', 'P', 'phase', 'Hvap', 'T_limit',
                     'price_kJ', 'price_kmol', 'efficiency')
        def __init__(self, species, molfrac, T, P, phase, Hvap, T_limit,
                     price_kJ, price_kmol, efficiency):
            self.species = species
            self.molfrac = molfrac
            self.T = T
            self.P = P
            self.phase = phase
            self.Hvap = Hvap
            self.T_limit = T_limit
            self.price_kJ = price_kJ
            self.price_kmol = price_kmol
            self.efficiency = efficiency
        
        def __repr__(self):
            return f"<{type(self).__name__}: T={self.T} K>"
        
        def _ipython_display_(self):
            print(f"{type(self).__name__}:\n"
                 +f" species               {repr(self.species)}\n"
                 +f" molfrac               ({', '.join([format(i,'.2f') for i in self.molfrac])},)\n"
                 +f" T           K         {self.T:,.2f}\n"
                 +f" P           Pa        {self.P:,.0f}\n"
                 +f" phase                 '{self.phase}'\n"
                 +f" Hvap        kJ/kmol   {self.Hvap and format(self.Hvap, ',.4g')}\n"
                 +f" T_limit     K         {self.T_limit and format(self.T_limit, ',.2f')}\n"
                 +f" price_kJ    USD/kJ    {self.price_kJ:.4g}\n"
                 +f" price_kmol  USD/kmol  {self.price_kmol:.4g}\n"
                 +f" efficiency            {self.efficiency:.2f}")
            
        show = _ipython_display_
            

    cooling_agents = {
        'Cooling water': UtilityAgent(_Water, (1,), 305.372, 101325, 'l',
                                      None, 324.817, 0, 4.8785e-4, 1),
        'Chilled water': UtilityAgent(_Water, (1,), 280.372, 101325, 'l',
                                      None, 300.372, -5e-6, 0, 1),
        'Chilled Brine': UtilityAgent(_Water, (1,), 255.372, 101325, 'l',
                                      None, 275.372, -8.145e-6, 0, 1)}

    heating_agents = {
        'Low pressure steam': UtilityAgent(_Water, (1,), 411.494, 344738.0, 'g',
                                           3.89e+04, None, 0, 0.2378, 0.95),
        'Medium pressure steam': UtilityAgent(_Water, (1,), 454.484, 1034214.0,
                                              'g', 3.63e+04, None, 0, 0.2756, 0.90),
        'High pressure steam': UtilityAgent(_Water, (1,), 508.858, 3102642.0, 'g',
                                            3.21e+04, None, 0, 0.3171, 0.85)}

    def __init__(self, efficiency=None):
        self.ID = ''
        self.cost = self.flow = self.duty = 0
        self.efficiency = efficiency
        
        #: tuple[bool, float] Cached arguments:
        #:     * [0]: [bool] True if duty is negative 
        #:     * [1]: [float] Temperature of operation
        self._args = None

    def _init_streams(self, flow, species, T, P, phase):
        """Initialize utility streams."""
        self._fresh = Stream(None, flow, species, T=T, P=P, phase=phase)
        self._used = Stream.proxy(None, self._fresh)

    def __call__(self, duty, T_in, T_out=None):
        """Calculate utility requirements given the essential parameters.
        
        **Parameters**
        
            **duty:** [float] Unit duty requirement (kJ/hr)
            
            **T_in:** [float] Entering process stream temperature (K)
            
            **T_out:** [float] Exit process stream temperature (K)
        
        """
        # Set pinch and operating temperature
        if T_out is None:
            T_pinch = T_op = T_in
        else:
            T_pinch, T_op = self._get_pinch(duty, T_in, T_out)
        
        if duty == 0:
            self.ID = ''
            self.flow = self.duty = self.cost = 0
            self._args = None
            return
        
        negduty = duty < 0
        args = (negduty, T_op)
        if self._args == args:
            if negduty:
                agent = self.cooling_agents[self.ID]
            else:
                agent = self.heating_agents[self.ID]
        else:
            self._args = args
            # Select heat transfer agent
            if negduty:
                agent = self._select_cooling_agent(T_op)
            else:
                agent = self._select_heating_agent(T_op)
            
        # Calculate utility flow rate requirement
        efficiency = self.efficiency or agent.efficiency
        duty = duty/efficiency
        if agent.Hvap:
            self._update_flow_wt_phase_change(duty, agent.Hvap)
        else:
            self._update_flow_wt_pinch_T(duty, T_pinch, agent.T_limit, negduty)
        
        # Update and return results
        self.flow = mol = self._fresh.molnet
        self.duty = duty
        self.cost = agent.price_kJ*duty + agent.price_kmol*mol

    @staticmethod
    def _get_pinch(duty, T_in, T_out):
        """Return pinch temperature and operating temperature."""
        if duty < 0:
            return (T_in, T_out) if T_in > T_out else (T_out, T_in)
        else:
            return (T_in, T_out) if T_in < T_out else (T_out, T_in)
    
    # Selection of a heat transfer agent
    def _select_cooling_agent(self, T_pinch):
        """Select a cooling agent that works at the pinch temperature and return relevant information.
        
        **Parameters**

             **T_pinch:**  [float] Pinch temperature of process stream (K)
        
        **Returns:** [UtilityAgent]
        
        """
        dt = 2*self.dT
        T_max = T_pinch - dt
        for ID, agent in self.cooling_agents.items():
            if T_max > agent.T:
                if self.ID != ID:
                    self._init_streams(agent.molfrac, agent.species,
                                       agent.T, agent.P, agent.phase)
                    self.ID = ID
                return agent
        raise RuntimeError(f'no cooling agent that can cool under {T_pinch} K')
            
    def _select_heating_agent(self, T_pinch):
        """Return a heating agent that works at the pinch temperature and return relevant information.
        
        **Parameters**

             **T_pinch:**  [float] Pinch temperature of process stream (K)
        
        **Returns:** [UtilityAgent]
        
        """
        dt = 2*self.dT
        T_min = T_pinch + dt
        for ID, agent in self.heating_agents.items():
            if T_min < agent.T:
                if self.ID != ID:
                    self._init_streams(agent.molfrac, agent.species,
                                       agent.T, agent.P, agent.phase)
                    self.ID = ID
                return agent
        raise RuntimeError(f'no heating agent that can heat over {T_pinch} K')

    # Main Calculations
    def _update_flow_wt_pinch_T(self, duty, T_pinch, T_limit, negduty):
        """Set utility Temperature at the pinch, calculate and set minimum net flowrate of the utility to satisfy duty and update."""
        self._used.T = self._T_exit(T_pinch, self.dT, T_limit, negduty)
        self._update_utility_flow(self._fresh, self._used, duty)

    def _update_flow_wt_phase_change(self, duty, latent_heat):
        """Change phase of utility, calculate and set minimum net flowrate of the utility to satisfy duty and update."""
        f = self._fresh
        u = self._used
        u._phase = 'g' if f._phase=='l' else 'l'
        u._mol[:] = duty/latent_heat

    # Subcalculations
    @staticmethod
    def _update_utility_flow(fresh, utility, duty):
        """Changes flow rate of utility such that it can meet the duty requirement"""
        utility._mol *= duty/(fresh.H - utility.H)

    @staticmethod
    def _T_exit(T_pinch, dT, T_limit, negduty):
        """Return exit temperature of the utility in a counter current heat exchanger

        **Parameters**

             **T_pinch:** [float] Pinch temperature of process stream (K)

             **dT:** [float] Pinch temperature difference (K)

             **negduty:** [bool] True if exit temperature should be lower (process stream is gaining energy)

        """
        if negduty:
            T_exit = T_pinch - dT
            if T_limit and T_limit < T_exit: T_exit = T_limit
        else:
            T_exit = T_pinch + dT
            if T_limit and T_limit > T_exit: T_exit = T_limit
        return T_exit

    def _info_data(self, duty, flow, cost):
        # Get units of measure
        su = self.display_units
        units = self._units
        duty_units = duty or su.duty
        flow_units = flow or su.flow
        cost_units = cost or su.cost
        
        # Select flow dimensionality
        flow_dim = _Q(0, flow_units).dimensionality
        if flow_dim == mol_flow_dim:
            flowattr = 'molnet'
        elif flow_dim == mass_flow_dim:
            flowattr = 'massnet'
        elif flow_dim == vol_flow_dim:
            flowattr = 'volnet'
        else:
            raise DimensionError(f"dimensions for flow units must be in molar, mass or volumetric flow rates, not '{flow_dim}'")
        
        # Change units and return info string
        try:
            u_in = self._fresh
            flownet = getattr(u_in, flowattr)
            flowunits = u_in.units[flowattr]
            flow = _Q(flownet, flowunits).to(flow_units).magnitude
        except:
            flow = _Q(self.flow, 'kmol/hr').to(flow_units).magnitude
        
        duty = _Q(self.duty, units['duty']).to(duty_units).magnitude
        cost = _Q(self.cost, units['cost']).to(cost_units).magnitude
        return duty, flow, cost, duty_units, flow_units, cost_units
        
    def __repr__(self):
        if self.ID:
            duty, flow, cost, duty_units, flow_units, cost_units = self._info_data(None, None, None)
            return f'<{self.ID}: {self.duty:.3g} {duty_units}, {self.flow:.3g} {flow_units}, {self.cost:.3g} {cost_units}>'
        else:
            return f'<{type(self).__name__}: None>'
        
    # Representation
    def _info(self, duty, flow, cost):
        """Return string related to specifications"""
        if not self.ID:
            return (f'{type(self).__name__}: None\n'
                    +f' duty: 0\n'
                    +f' flow: 0\n'
                    +f' cost: 0')
        else:
            (duty, flow, cost, duty_units,
             flow_units, cost_units) = self._info_data(duty, flow, cost)
            return (f'{type(self).__name__}: {self.ID}\n'
                    +f' duty:{duty: .3g} {duty_units}\n'
                    +f' flow:{flow: .3g} {flow_units}\n'
                    +f' cost:{cost: .3g} {cost_units}')
            

    def show(self, duty=None, flow=None, cost=None):
        """Print all specifications"""
        print(self._info(duty, flow, cost))
    _ipython_display_ = show


del _Water